'''
PSF interpolation for CSST-Sim

NOTE: [iccd, iwave, ipsf] are counted from 1 to n, but [tccd, twave, tpsf] are counted from 0 to n-1
'''

import galsim
import numpy as np

import os
import time
import copy
import PSF.PSFInterp.PSFConfig as myConfig
import PSF.PSFInterp.PSFUtil as myUtil
from PSF.PSFModel import PSFModel

LOG_DEBUG = False  #***#
NPSF      = 900    #***# 30*30
iccdTest  = 1      #***#


class PSFInterp(PSFModel):
    # def __init__(self, PSF_data=None, params=None, PSF_data_file=None):
    def __init__(self, chip, PSF_data=None, PSF_data_file=None, sigSpin=0, psfRa=0.15):
        """
        The PSF data matrix is either given by a object parameter or read in from a file.
        Parameters:
            PSF_data: The PSF data matrix object
            params: Other parameters?
            PSF_data_file: The file for PSF data matrix (optional).
        """
        if LOG_DEBUG:
            print('===================================================')
            print('DEBUG: psf module for csstSim ' \
                  +time.strftime("(%Y-%m-%d %H:%M:%S)", time.localtime()), flush=True)
            print('===================================================')

        self.sigSpin = sigSpin
        self.sigGauss = psfRa

        self.iccd = int(chip.getChipLabel(chipID=chip.chipID))
        if PSF_data_file == None:
            PSF_data_file = '/data/simudata/CSSOSDataProductsSims/data/csstPSFdata/CSSOS_psf_20210108/CSST_psf_ciomp_2p5um_cycle3_ccr90'
        self.nwave= self._getPSFwave(self.iccd, PSF_data_file)
        self.PSF_data = self._loadPSF(self.iccd, PSF_data_file)

        if LOG_DEBUG:
            print('nwave-{:} on ccd-{:}::'.format(self.nwave, self.iccd), flush=True)
            print('self.PSF_data ... ok', flush=True)
            print('Preparing self.[psfMat,cen_col,cen_row] for psfMaker ... ', end='', flush=True)

        ngy, ngx    = self.PSF_data[0][0]['psfMat'].shape
        self.psfMat = np.zeros([self.nwave, NPSF, ngy, ngx], dtype=np.float32)
        self.cen_col= np.zeros([self.nwave, NPSF], dtype=np.float32)
        self.cen_row= np.zeros([self.nwave, NPSF], dtype=np.float32)
        self.hoc    =[]
        self.hoclist=[]
        for twave in range(self.nwave):
            for tpsf in range(NPSF):
                self.psfMat[twave, tpsf, :, :] = self.PSF_data[twave][tpsf]['psfMat']
                self.PSF_data[twave][tpsf]['psfMat'] = 0 ###free psfMat

                self.pixsize = self.PSF_data[twave][tpsf]['pixsize']*1e-3 ##mm
                self.cen_col[twave, tpsf] = self.PSF_data[twave][tpsf]['image_x'] +0.*self.pixsize + self.PSF_data[twave][tpsf]['centroid_x']
                self.cen_row[twave, tpsf] = self.PSF_data[twave][tpsf]['image_y'] +0.*self.pixsize + self.PSF_data[twave][tpsf]['centroid_y']

            #hoclist on twave for neighborsFinding
            hoc,hoclist = myUtil.findNeighbors_hoclist(self.cen_col[twave], self.cen_row[twave])
            self.hoc.append(hoc)
            self.hoclist.append(hoclist)

        if LOG_DEBUG:
            print('ok', flush=True)


    def _getPSFwave(self, iccd, PSF_data_file):
        """
        Get # of sampling waves on iccd
        Parameters:
            iccd: The chip of i-th ccd
            PSF_data_file: The file for PSF data matrix
        Returns:
            nwave: The number of the sampling waves
        """
        strs = os.listdir(PSF_data_file + '/ccd{:}'.format(iccd))
        nwave = 0
        for _ in strs:
            if 'wave_' in _:
                nwave += 1
        return nwave



    def _loadPSF(self, iccd, PSF_data_file):
        """
        load psf-matrix on iccd
        Parameters:
            iccd: The chip of i-th ccd
            PSF_data_file: The file for PSF data matrix
        Returns:
            psfSet: The matrix of the csst-psf
        """
        psfSet = []
        for ii in range(self.nwave):
            iwave = ii+1
            if LOG_DEBUG:
                print('self._loadPSF: iwave::', iwave, flush=True)
            psfWave = []
            for jj in range(NPSF):
                ipsf = jj+1
                psfInfo = myConfig.LoadPSF(iccd, iwave, ipsf, PSF_data_file, InputMaxPixelPos=True, PSFCentroidWgt=True, PSFBinning=True, PSFConvGauss=True)
                psfWave.append(psfInfo)
            psfSet.append(psfWave)
        if LOG_DEBUG:
            print('psfSet has been loaded:', flush=True)
            print('psfSet[iwave][ipsf][keys]:', psfSet[0][0].keys(), flush=True)
        return psfSet


    
    def _preprocessPSF(self):
        pass


    def _findWave(self, bandpass):
        for twave in range(self.nwave):
            bandwave = self.PSF_data[twave][0]['wavelength']
            if bandpass.blue_limit < bandwave and bandwave < bandpass.red_limit:
                return twave
        return -1
 


    def get_PSF(self, chip, pos_img, bandpass, pixSize=0.0184, galsimGSObject=True, findNeighMode='treeFind', folding_threshold=5.e-3):
        """
        Get the PSF at a given image position

        Parameters:
            chip: A 'Chip' object representing the chip we want to extract PSF from.
            pos_img: A 'galsim.Position' object representing the image position.
            bandpass: A 'galsim.Bandpass' object representing the wavelength range.
            pixSize: The pixels size of psf matrix
            findNeighMode: 'treeFind' or 'hoclistFind'
        Returns:
            PSF: A 'galsim.GSObject'.
        """

        pixSize = np.rad2deg(self.pixsize*1e-3/28)*3600  #set psf pixsize

        #***#  assert self.iccd == chip.chipID, 'ERROR: self.iccd != chip.chipID'
        # twave = bandpass-1  #***# ??? #self.findWave(bandpass)  ###twave=iwave-1 as that in NOTE
        assert self.iccd == int(chip.getChipLabel(chipID=chip.chipID)), 'ERROR: self.iccd != chip.chipID'
        twave = self._findWave(bandpass)
        if twave == -1:
            print("!!!PSF bandpass does not match.")
            exit()
        PSFMat = self.psfMat[twave]
        cen_col= self.cen_col[twave]
        cen_row= self.cen_row[twave]

        # px = pos_img[0]
        # py = pos_img[1]
        px = (pos_img.x - chip.cen_pix_x)*0.01
        py = (pos_img.y - chip.cen_pix_y)*0.01
        if findNeighMode == 'treeFind':
            imPSF = myConfig.psfMaker_IDW(px, py, PSFMat, cen_col, cen_row, IDWindex=2, OnlyNeighbors=True, PSFCentroidWgt=True)
        if findNeighMode == 'hoclistFind':
            imPSF = myConfig.psfMaker_IDW(px, py, PSFMat, cen_col, cen_row, IDWindex=2, OnlyNeighbors=True, hoc=self.hoc[twave], hoclist=self.hoclist[twave], PSFCentroidWgt=True)

        if galsimGSObject:
            img = galsim.ImageF(imPSF, scale=pixSize)
            gsp = galsim.GSParams(folding_threshold=folding_threshold)
            self.psf = galsim.InterpolatedImage(img, gsparams=gsp)
            # gaussPSF = galsim.Gaussian(sigma=0.04, gsparams=gsp)
            # self.psf    = galsim.Convolve(gaussPSF, self.psf)
            return self.PSFspin(x=px/0.01, y=py/0.01)
        return imPSF

    def PSFspin(self, x, y):
        """
        The PSF profile at a given image position relative to the axis center

        Parameters:
        theta : spin angles in a given exposure in unit of [arcsecond]
        dx, dy: relative position to the axis center in unit of [pixels]

        Return:
        Spinned PSF: g1, g2 and axis ratio 'a/b'
        """
        a2Rad = np.pi/(60.0*60.0*180.0)
        
        ff = self.sigGauss * 0.107 * (1000.0/10.0) # in unit of [pixels]
        rc = np.sqrt(x*x + y*y)
        cpix = rc*(self.sigSpin*a2Rad)

        beta = (np.arctan2(y,x) + np.pi/2)
        ell = cpix**2/(2.0*ff**2+cpix**2)
        qr = np.sqrt((1.0+ell)/(1.0-ell))
        PSFshear = galsim.Shear(e=ell, beta=beta*galsim.radians)
        return self.psf.shear(PSFshear), PSFshear




if __name__ == '__main__':
    if True:
        psfPath = '/data/simudata/CSSOSDataProductsSims/data/csstPSFdata/CSSOS_psf_20210108/CSST_psf_ciomp_2p5um_cycle3_ccr90'
        psfCSST = PSFInterp(PSF_data_file = psfPath)
        iwave= 1
        ipsf = 665
        pos_img = [psfCSST.cen_col[iwave, ipsf], psfCSST.cen_row[iwave, ipsf]]
        img = psfCSST.get_PSF(1, pos_img, iwave, galsimGSObject=True)
        print('haha')

    
    if False:
        #old version (discarded)
        #plot check-1
        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=(18,5))
        ax = plt.subplot(1,3,1)
        plt.imshow(img)
        plt.colorbar()
        ax = plt.subplot(1,3,2)
        imgx = psfCSST.itpPSF_data[iwave][ipsf]['psfMat']
        imgx/= np.sum(imgx)
        plt.imshow(imgx)
        plt.colorbar()
        ax = plt.subplot(1,3,3)
        plt.imshow(img - imgx)
        plt.colorbar()
        plt.savefig('test/figs/test1.jpg')

    if False:
        #old version (discarded)
        #plot check-2: 注意图像坐标和全局坐标
        fig = plt.figure(figsize=(8,8), dpi = 200)
        img = psfCSST.PSF_data[iwave][ipsf]['psfMat']
        npix = img.shape[0]
        dng  = 105
        imgg = img[dng:-dng, dng:-dng]
        plt.imshow(imgg)
        imgX = psfCSST.PSF_data[iwave][ipsf]['image_x']  #in mm
        imgY = psfCSST.PSF_data[iwave][ipsf]['image_y']  #in mm
        deltX= psfCSST.PSF_data[iwave][ipsf]['centroid_x'] #in mm
        deltY= psfCSST.PSF_data[iwave][ipsf]['centroid_y'] #in mm
        maxX = psfCSST.PSF_data[iwave][ipsf]['max_x']
        maxY = psfCSST.PSF_data[iwave][ipsf]['max_y']
        cenPix_X = npix/2 + deltX/0.005
        cenPix_Y = npix/2 - deltY/0.005
        maxPix_X = npix/2 + maxX/0.005-1
        maxPix_Y = npix/2 - maxY/0.005-1
        plt.plot([cenPix_X-dng],[cenPix_Y-dng], 'rx', ms = 20)
        plt.plot([maxPix_X-dng],[maxPix_Y-dng], 'b+', ms=20)

        from scipy import ndimage
        y, x = ndimage.center_of_mass(img)
        plt.plot([x-dng],[y-dng], 'rx', ms = 10, mew=4)
        x, y = myUtil.findMaxPix(img)
        plt.plot([x-dng],[y-dng], 'b+', ms = 10, mew=4)
        plt.savefig('test/figs/test2.jpg')

